from meloetta.frameworks.mewzero.model.heads.action_type_head import ActionTypeHead
from meloetta.frameworks.mewzero.model.heads.max_move_head import MaxMoveHead
from meloetta.frameworks.mewzero.model.heads.move_head import MoveHead
from meloetta.frameworks.mewzero.model.heads.switch_head import SwitchHead
from meloetta.frameworks.mewzero.model.heads.flag_head import FlagsHead
